%% Computational cost as a function of the error
%% stochastic Maxwell equation
%% exponential integrator

clear all; close all
rng(111)

fprintf('computcostEXP.m')

method='Exp';

% parameter in front of the noise
% yesno=0 -> no noise
yesno=1.; 

% time discretisation
T=.5; N=2^15; dt=T/N;
tt=[0+dt:dt:T];

% space discretisation wlog Mx=My and dx=dy
Il=0;Ir=1; % x interval
M=2^4; % 2^4 takes a lot of time  
dx=(Ir-Il)/(M); h_ex=dx;
xx=dx*[1:M-1]'; 
xx12=[Il+dx/2:dx:Ir-dx/2];
Ily=0;Iry=1; % y interval
My=M; 
dy=(Iry-Ily)/(My);
yy=dy*[1:My-1]'; 
yy12=[Il+dy/2:dy:Ir-dy/2];

% FD matrices d/dx or d/dy assuming dx=dy !! 
AEy=sparse(M*(M-1),(M-1)^2);
AEx=AEy;
for j=1:(M-1)^2
  AEy(j,j)=1;
end
for j=1:(M-1)^2
  AEy(j+(M-1),j)=-1;
end
AA=spdiags([-1*ones(M-1,1) ones(M-1,1)],[-1:0],M,M-1);
Ix=speye(M-1); 
AEx=kron(Ix,AA);
AHy=-AEy.';
AHx=-AEx.';
zorro1=spalloc((M-1)^2,(M-1)^2,0);
zorro2=spalloc(M*(M-1),M*(M-1),0);
% approximation of matrix [0 -d/dy d/dx; -d/dy 0 0; d_x 0 0]
B=[ zorro1  -AHy +AHx; -AEy zorro2 zorro2; AEx zorro2 zorro2 ]; 
% same dx and dy
B=B/dx; 

% initial conditions
E3=zeros(M-1,M-1); 
H1=zeros(M-1,M); 
H2=zeros(M,M-1);
% E3 must be zero on the boundary
for j=1:M-1
  for k=1:M-1
      %% pulse
      E3(j,k)=1.1*exp(-50*((xx(j)-.5).^2+(yy(k)-.5).^2)); 
      %% chz16.pdf
      %E3(j,k)=sin(2*pi*xx(j))*sin(2*pi*yy(k))*.3;
      %% rdm
      %E3(j,k)=randn*sin(2*pi*xx(j))*sin(2*pi*yy(k));
   end
end   
E3(end,:)=0.*E3(end,:); E3(1,:)=0.*E3(1,:);
E3(:,1)=0.*E3(:,1); E3(:,end)=0.*E3(:,end); 

% H1 is constant in x
for j=1:M-1
  H1(j,:)=ones(M,1)*randn;
end  
%H1

% H2 is constant in y
for k=1:M-1
  H2(:,k)=ones(M,1)*randn;
end  
%H2

Uinit=[E3(:); H1(:); H2(:)];

% eigenvalues for the noise from chz16.pdf more or less
E3_tempo=zeros(M-1);
for j=1:M-1
  for k=1:M-1
      E3_tempo(j,k)=3/(j^3+k^3);
   end
end   
%  E3_tempo(end,:)=0.*E3_tempo(end,:); E3_tempo(1,:)=0.*E3_tempo(1,:);
%  E3_tempo(:,1)=0.*E3_tempo(:,1); E3_tempo(:,end)=0.*E3_tempo(:,end); 

for j=1:M-1
  for k=0:M-1
      H1_tempo(j,k+1)=3/(j^3+(k+1/2)^3);
   end
end   
for j=0:M-1
  for k=1:M-1
      H2_tempo(j+1,k)=3/((j+1/2)^3+k^3);
   end
end   
E3tempo=reshape(E3_tempo,1,(M-1)^2);
H1tempo=reshape(H1_tempo,1,M*(M-1));
H2tempo=reshape(H2_tempo,1,M*(M-1));
E_ex=[1.*E3tempo 1*H1tempo 1*H2tempo]'; 

% number of samples
MS=100;

for mm=1:MS;

  %% noise
  Wtt=sqrt(dt)*randn(N,(M-1)^2+2*M*(M-1));
  W=zeros(N,(M-1)^2+2*M*(M-1));
  for j=1:(M-1)^2+2*M*(M-1)
    W(:,j)=E_ex(j)*Wtt(:,j);
  end
  
  %% initial value
  U1=Uinit; 
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%% Calcul de la solution de reference %%%
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  disp(strcat('Calcul de la solution de reference pour le tirage numero ',num2str(mm)));  

   %% exponential integrator
   if method=='Exp'
      solref=U1;
      
      expA=expm(dt*B); 
    
      V=zeros(N,(M-1)^2+2*M*(M-1));
      for i=1:N
        V(i,:)=sum(W((i-1)+1:i,:),1);
      end  
      Wn=yesno*V;
        
      for j=1:N
        %% StochExp 
        U1=expA*( U1+Wn(j,:)'+ dt*(U1) );
        E3tsexp=U1(1:(M-1)^2); 
        E3tsexp(end,:)=0.*E3tsexp(end,:); E3tsexp(1,:)=0.*E3tsexp(1,:);
        E3tsexp(:,1)=0.*E3tsexp(:,1); E3tsexp(:,end)=0.*E3tsexp(:,end); 
        U1(1:(M-1)^2)=E3tsexp;
        solref=U1;
     end;
  end
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%% Calcul des solutions approchees %%%
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  untousles=2.^[1 2 3 4 5 6 7 8];
  for numpasdetemps=1:length(untousles)
      U1=Uinit; 
      
      facteurraff=untousles(numpasdetemps)
      Ntloc=N/facteurraff;
      dtloc=dt*facteurraff;
      dtloc=(T-0)/Ntloc;
      
      %%  exponential integrator
      if method=='Exp'
          tdeb=cputime;
          
          expA=expm(dtloc*B);
          
          for t=1:Ntloc
            nnoise=yesno*sum(W(facteurraff*(t-1)+1:facteurraff*t,:),1);
            %% StochExp 
            U1=expA*( U1+ nnoise' + dtloc*(U1) ); 
            E3tsexp=U1(1:(M-1)^2); 
            E3tsexp(end,:)=0.*E3tsexp(end,:); E3tsexp(1,:)=0.*E3tsexp(1,:);
            E3tsexp(:,1)=0.*E3tsexp(:,1); E3tsexp(:,end)=0.*E3tsexp(:,end); 
            U1(1:(M-1)^2)=E3tsexp; 
          end;
          tfin=cputime;
          erreurL2carreefinaletirage(mm,numpasdetemps)=sum( abs(U1-solref).^2/length(U1)^2 );
          tempscalcultirage(mm,numpasdetemps)=double(tfin-tdeb);
      end;
        
  end; % fin de la boucle sur les pas de temps pour les solutions approchees
  
end; % fin de la boucle sur les trajectoires

for numpasdetemps=1:length(untousles)
    erreurL2carreefinalemoyenne(numpasdetemps) = sum(erreurL2carreefinaletirage(:,numpasdetemps))/MS;
    tempcalcul(numpasdetemps)=sum(tempscalcultirage(:,numpasdetemps));
end;

figure();
XX=log10((T-0)/N.*untousles);
YY=log10(erreurL2carreefinalemoyenne);
plot(XX,YY);
xlabel(strcat('time step (log10-scale)'));
ylabel(strcat('averaged squared L2 final error over ',num2str(MS),' samples (log10-scale) for the ',method,' method'));
title(strcat('Numerical efficiency of the ',method,' method'));
polyfit(XX,YY,1)

save('Exp-err.mat','XX','YY');

figure();
XX=log10(erreurL2carreefinalemoyenne);
YY=log10(tempcalcul);
plot(XX,YY);
xlabel(strcat('averaged squared L2 final error over ',num2str(MS),' samples (log10-scale)'));
ylabel('total computational time in seconds (log10-scale)');
title(strcat('Numerical efficiency of the ',method,' method'));

save('Exp-computcost','XX','YY');




