%% seems 
%% ms-errors in time 
%% Stagged uniform grid
%% PEC+PMC 
%% Parameters similar to Chen, Hong, Zhang 2016
%% Stochastic exponential method, BEM, EM
%% F(V)=V+cos(V), G(V)=yesno*sin(V), expected rate 1/2

clear all; close all;
rng(93)

fprintf('maxwell2dNL.m\n');

% parameter in front of the noise
% yesno=0 -> no noise
yesno=1; 

% number of samples
MM=500

% time discretisation
t_0=0; t_end=.5;
dt_num=2.^(-8:-1:-13);
dt_ex=dt_num(end); 

% space discretisation wlog Mx=My and dx=dy
Il=0;Ir=1; % x interval
M=2^4; % 2^4 takes a lot of time  
dx=(Ir-Il)/(M); h_ex=dx;
xx=dx*[1:M-1]'; 
xx12=[Il+dx/2:dx:Ir-dx/2];
Ily=0;Iry=1; % y interval
My=M; 
dy=(Iry-Ily)/(My);
yy=dy*[1:My-1]'; 
yy12=[Il+dy/2:dy:Ir-dy/2];

% FD matrices d/dx or d/dy assuming dx=dy !! 
AEy=sparse(M*(M-1),(M-1)^2);
AEx=AEy;
for j=1:(M-1)^2
  AEy(j,j)=1;
end
for j=1:(M-1)^2
  AEy(j+(M-1),j)=-1;
end
AA=spdiags([-1*ones(M-1,1) ones(M-1,1)],[-1:0],M,M-1);
Ix=speye(M-1); 
AEx=kron(Ix,AA);
AHy=-AEy.';
AHx=-AEx.';
zorro1=spalloc((M-1)^2,(M-1)^2,0);
zorro2=spalloc(M*(M-1),M*(M-1),0);
% approximation of matrix [0 -d/dy d/dx; -d/dy 0 0; d_x 0 0]
B=[ zorro1  -AHy +AHx; -AEy zorro2 zorro2; AEx zorro2 zorro2 ]; 
% same dx and dy
B=B/dx; 

% mesh and stepsizes
Nt_num=floor((t_end-t_0)./dt_num);
Nt_ex=floor((t_end-t_0)/dt_ex);
h_num=h_ex*ones(size(dt_num,2),1)';

% initial conditions
E3=zeros(M-1,M-1); 
H1=zeros(M-1,M); 
H2=zeros(M,M-1);
% E3 must be zero on the boundary
for j=1:M-1
  for k=1:M-1
      %% pulse
      E3(j,k)=.1*exp(-50*((xx(j)-.5).^2+(yy(k)-.5).^2));
      %% chz16.pdf
      %E3(j,k)=sin(3*pi*xx(j))*sin(6*pi*yy(k))*.3;
      %% rdm
      %E3(j,k)=randn*sin(2*pi*xx(j))*sin(2*pi*yy(k));
   end
end   

% H1 is constant in x
for j=1:M-1
  H1(j,:)=ones(M,1)*randn;
end  
%H1

% H2 is constant in y
for k=1:M-1
  H2(:,k)=ones(M,1)*randn;
end  
%H2

Uinit=[E3(:); H1(:); H2(:)];

USEXP=zeros(length(dt_num),length(Uinit));
UEM=zeros(length(dt_num),length(Uinit));
UBEM=zeros(length(dt_num),length(Uinit));

eSEXP=zeros(length(dt_num),length(Uinit),MM);
eEM=zeros(length(dt_num),length(Uinit),MM);
eBEM=zeros(length(dt_num),length(Uinit),MM);

% eigenvalues for the noise from chz16.pdf more or less
E3_tempo=zeros(M-1);
for j=1:M-1
  for k=1:M-1
      E3_tempo(j,k)=3/(j^3+k^3);
   end
end   
for j=1:M-1
  for k=0:M-1
      H1_tempo(j,k+1)=3/(j^3+(k+1/2)^3);
   end
end   
for j=0:M-1
  for k=1:M-1
      H2_tempo(j+1,k)=3/((j+1/2)^3+k^3);
   end
end   
E3tempo=reshape(E3_tempo,1,(M-1)^2);
H1tempo=reshape(H1_tempo,1,M*(M-1));
H2tempo=reshape(H2_tempo,1,M*(M-1));
E_ex=[1.*E3tempo 1*H1tempo 1*H2tempo]'; 

for m = 1:MM
 
  %% noise
  Wtt=sqrt(dt_ex)*randn((M-1)^2+2*M*(M-1),Nt_ex);
  W=zeros((M-1)^2+2*M*(M-1),Nt_ex);
  for k1=1:Nt_ex
    for j=1:(M-1)^2+2*M*(M-1)
      W(j,k1)=E_ex(j)*Wtt(j,k1);
    end
  end
    
  for j=1:length(dt_num)
    Nt=(t_end-t_0)/dt_num(j);
    dt=dt_num(j);
    k=dt/dt_ex;
    
    expA=expm(dt*B); 
    Bbem=speye((M-1)^2+2*M*(M-1))-dt*B;
    
    %% initial value
    U1=Uinit; U1em=Uinit; U1bem=Uinit;
    for t=1:Nt
      Wn=sum(W(:,k*(t-1)+1:k*t),2);
      %% StochExp 
      U1=expA*( U1 + dt*U1 +dt*cos(U1)+ sin(U1).*Wn ); 
      %% EM
      U1em=U1em+dt*B*U1em+ dt*U1em+dt*cos(U1em) + sin(U1em).*Wn;
      %% BEM
      U1bem=Bbem\( U1bem+ dt*U1bem+dt*cos(U1bem)+ sin(U1bem).*Wn );
    end % t
    U(j,m,:)=U1; 
    Uem(j,m,:)=U1em;
    Ubem(j,m,:)=U1bem;   
  end  %j     
  m
end % m 

XerrSEXP=zeros(MM,length(dt_num)-1);
XerrEM=zeros(MM,length(dt_num)-1);
XerrBEM=zeros(MM,length(dt_num)-1);
for path=1:MM
    for i=1:length(dt_num)-1
        XerrSEXP(path,i)=sum((    U(i,path,:)- U(length(dt_num),path,:) ).^2)*dx^2;% strong order
        XerrEM(path,i)=sum((   Uem(i,path,:)- U(length(dt_num),path,:) ).^2)*dx^2;% strong order
        XerrBEM(path,i)=sum((   Ubem(i,path,:)- U(length(dt_num),path,:) ).^2)*dx^2;% strong order
    end
end

meanXerrSEXP=sqrt(mean(XerrSEXP,1)); %strong order
meanXerrEM=sqrt(mean(XerrEM,1));
meanXerrBEM=sqrt(mean(XerrBEM,1));

set(0,'DefaultTextFontSize',12)
set(0,'DefaultAxesFontSize',12)
set(0,'DefaultLineLineWidth',2)
set(0,'DefaultLineMarkerSize',10)

figure(),
loglog(dt_num(1:end-1),dt_num(1:end-1).^(.5),'k--', ...
       dt_num(1:end-1),dt_num(1:end-1).^1,'k--', ...
       dt_num(1:end-1),meanXerrSEXP,'ks-',dt_num(1:end-1), ... 
       meanXerrEM,'kd-',dt_num(1:end-1),meanXerrBEM,'k*-');
%axis([(dt_num(end-1)) (dt_num(1)) 10^-10 10^0])
xlabel('$\Delta t$','Interpreter','latex','FontSize',14);
ylabel('MS-Error','Rotation',0,'HorizontalAlignment','right','FontSize',14);
legend('Slope 1/2','Slope 1','Exp','EM','SEM','Location','SouthEast');
str='$M_s=$';
title(sprintf('%s %d',str,MM),'Interpreter','latex','FontSize',14);

print -depsc2 -r0 maxwell2dNL.eps

save('maxwell2dNL.mat','dt_num','meanXerrBEM', 'meanXerrEM', 'meanXerrSEXP') 

